"""Make pretty footprint plots.

This file contains plotting code by Geert.
Alternative plotting routines from Fergal are available.
"""
import sys
import numpy as np

from . import getKeplerFov, logger

# Now try loading matplotlib
try:
    import matplotlib.pyplot as pl
    params = {
                'axes.linewidth': 1.5,
                'axes.labelsize': 24,
                'font.family': 'sans-serif',
                'font.size': 22,
                'legend.fontsize': 14,
                'xtick.labelsize': 16,
                'ytick.labelsize': 16,
                'text.usetex': False,
             }
    pl.rcParams.update(params)
except Exception:
    logger.error('This feature requires matplotlib to be installed.')
    sys.exit(1)


class K2FootprintPlot(object):

    def __init__(self, axes=None, figsize=(16, 5)):
        if axes is None:
            self.fig = pl.figure(figsize=figsize)
            self.ax = self.fig.add_subplot(111)
        else:
            self.ax = axes
        self.ax.set_ylim([-35, 35])
        self.ax.set_xlim([0, 360])
        self.ax.set_xlabel("RA [deg]")
        self.ax.set_ylabel("Dec [deg]")
        try:
            self.fig.tight_layout()
        except AttributeError:  # We didn't create a fig above
            pass

    def plot_campaigns(self, campaigns=18):
        """Plot the outlines of all campaigns."""
        for c in range(campaigns):
            self.plot_campaign_outline(c)

    def plot_campaign_outline(self, campaign=0, facecolor="#666666"):
        """Plot the outline of a campaign as a contiguous gray patch.

        Parameters
        ----------
        campaign : int
            K2 Campaign number.

        facecolor : str
            Color of the patch.
        """
        # The outline is composed of two filled rectangles,
        # defined by the first coordinate of the corner of four channels each
        fov = getKeplerFov(campaign)
        corners = fov.getCoordsOfChannelCorners()
        for rectangle in [[4, 75, 84, 11], [15, 56, 71, 32]]:
            ra_outline, dec_outline = [], []
            for channel in rectangle:
                idx = np.where(corners[::, 2] == channel)
                ra_outline.append(corners[idx, 3][0][0])
                dec_outline.append(corners[idx, 4][0][0])
            self.ax.fill(ra_outline + ra_outline[:1],
                         dec_outline + dec_outline[:1],
                         facecolor=facecolor, zorder=151, lw=0)
        # Print the campaign number on top of the outline
        ra, dec, roll = fov.getBoresight()
        self.ax.text(ra, dec, "{}".format(campaign),
                     fontsize=18, color="white", ha="center", va="center",
                     zorder=155)

    def plot_campaign(self, campaign=0, annotate_channels=True):
        """Plot all the active channels of a campaign."""
        fov = getKeplerFov(campaign)
        corners = fov.getCoordsOfChannelCorners()

        for ch in np.arange(1, 85, dtype=int):
            if ch in fov.brokenChannels:
                continue  # certain channel are no longer used
            idx = np.where(corners[::, 2] == ch)
            mdl = int(corners[idx, 0][0][0])
            out = int(corners[idx, 1][0][0])
            ra = corners[idx, 3][0]
            dec = corners[idx, 4][0]
            self.ax.fill(np.concatenate((ra, ra[:1])),
                         np.concatenate((dec, dec[:1])),
                         lw=0, facecolor="#bbbbbb", zorder=90)
            if annotate_channels:
                txt = "K2C{0}\n{1}.{2}\n#{3}".format(campaign, mdl, out, ch)
                self.ax.text(np.mean(ra), np.mean(dec), txt,
                             ha="center", va="center",
                             zorder=91, fontsize=10,
                             color="#666666", clip_on=True)

    def plot_ecliptic(self, size=100):
        try:
            from astropy.coordinates import SkyCoord
        except ImportError:
            logger.error("You need to install AstroPy for this feature.")
            return None
        try:
            icrs = SkyCoord(np.linspace(0, 359, num=size), 0,
                            unit="deg", frame="barycentrictrueecliptic").icrs
            self.ax.plot(icrs.ra, icrs.dec, lw=2, color="#666666")
        except ValueError:
            # only AstroPy 1.1 and up support ecliptic coordinates;
            # avoid crashing if an older version of AstroPy is at play
            pass

    def plot_galactic(self, size=150, color="#bbbbbb", textcolor="#777777"):
        try:
            from astropy.coordinates import SkyCoord
        except ImportError:
            logger.error("You need to install AstroPy for this feature.")
            return None
        icrs = SkyCoord(np.linspace(0, 359, num=size), 0,
                        unit="deg", frame="galactic").icrs
        self.ax.plot(icrs.ra, icrs.dec, lw=20, color=color)
        self.ax.text(116, -17, "Galactic Plane", rotation=-60,
                     fontsize=12, color=textcolor)

    def plot(self):
        self.plot_galactic()
        self.plot_ecliptic()
        self.plot_campaigns()


def create_context_plot(ra, dec, name="Your object"):
    """Creates a K2FootprintPlot showing a given position in context
    with respect to the campaigns."""
    plot = K2FootprintPlot()
    plot.plot_galactic()
    plot.plot_ecliptic()
    for c in range(0, 11):
        plot.plot_campaign_outline(c, facecolor="#666666")
    for c in [11, 12, 13, 14, 15, 16, 17, 18]:
        plot.plot_campaign_outline(c, facecolor="green")
    plot.ax.scatter(ra, dec, marker='x', s=250, lw=3, color="red", zorder=500)
    plot.ax.text(ra, dec - 2, name,
                 ha="center", va="top", color="red",
                 fontsize=20, fontweight='bold', zorder=501)
    return plot


def create_context_plot_zoomed(ra, dec, name="Your object", size=3):
    """Creates a K2FootprintPlot showing a given position in context
    with respect to the campaigns."""
    plot = K2FootprintPlot(figsize=(8, 8))
    for c in range(0, 19):
        plot.plot_campaign(c)
    plot.ax.scatter(ra, dec, marker='x', s=250, lw=3, color="red", zorder=500)
    plot.ax.text(ra, dec - 0.05*size, name,
                 ha="center", va="top", color="red",
                 fontsize=20, fontweight='bold', zorder=501)
    plot.ax.set_xlim([ra - size/2., ra + size/2.])
    plot.ax.set_ylim([dec - size/2., dec + size/2.])
    return plot


if __name__ == "__main__":
    plot = K2FootprintPlot()
    plot.plot_galactic()
    plot.plot_ecliptic()
    for c in range(0, 9):
        plot.plot_campaign_outline(c, facecolor="#666666")
    for c in [9, 10]:
        plot.plot_campaign_outline(c, facecolor="red")
    for c in [11, 12, 13, 14, 15, 16, 17, 18]:
        plot.plot_campaign_outline(c, facecolor="green")
    plot.fig.show()



"""
Construct various rotation matrices.

The documentation strings assume the "3-2-1" convention. A
vector should be rotated about the z-axis first, then the y-axis,
then finally the x-axis.

For ra and dec, rotate in ra first, then declination.

To rotate a vector, premultiply the the matrix returned
by these functions.

R = rotateInZ(angle)
vDash = np.dot(R, v)

To yaw, pitch and roll.
Y = yawLeftMat()
P = pitchUpMAt()
R = rollClockwiseMat()

rotMat = np.dot(R, P)
rotMat = np.dot(rotMat, Y)
vDash = np.dot(rotMat, v)

This is equivalent to vDash = R.P.Y.v

If you want to rotate the coordinate system instead of the vector
then you need to post multiply the matrix
vDash = np.dot(v, R), where vDash is the vector in the rotated
coord system.
"""
import numpy as np


def getAngleBetweenVectors(a, b, degrees=True):
    norm = np.linalg.norm

    cost = np.dot(a, b)
    cost /= norm(a) * norm(b)

    angle = np.arccos(cost)   # In radians
    if degrees:
        angle *= 180/np.pi
    return angle


def declinationRotationMatrix(theta_deg):
    """Rotate a vector in increasing declination

    Input:
    theta_deg   (float) Angle through which vectors should be
                rotated in degrees

    Returns:
    A matrix

    To rotate a vector, premultiply by this matrix.
    To rotate the coord sys underneath the vector, post multiply
    """
    return pitchUpMat(theta_deg)


def rightAscensionRotationMatrix(theta_deg):
    """Rotate a vector in increasing ra

    Input:
    theta_deg   (float) Angle through which vectors should be
                rotated in degrees

    Returns:
    A matrix

    To rotate a vector, premultiply by this matrix.
    To rotate the coord sys underneath the vector, post multiply
    """

    return yawLeftMat(theta_deg)


def yawLeftMat(theta_deg):
    """Thin wrapper to rotateInZ with a human readable name

    Input:
    theta_deg   (float) Angle through which vectors should be
                rotated in degrees

    Returns:
    A matrix

    To rotate a vector, premultiply by this matrix.
    To rotate the coord sys underneath the vector, post multiply
    """
    return rotateInZMat(theta_deg)


def pitchUpMat(theta_deg):
    """Thin wrapper to rotateInY with a human readable name

    Input:
    theta_deg   (float) Angle through which vectors should be
                rotated in degrees

    Returns:
    A matrix

    To rotate a vector, premultiply by this matrix.
    To rotate the coord sys underneath the vector, post multiply
    """
    return rotateInYMat(-theta_deg)


def rollClockwiseMat(theta_deg):
    """Thin wrapper to rotateInX with a human readable name

    Input:
    theta_deg   (float) Angle through which vectors should be
                rotated in degrees

    Returns:
    A matrix

    To rotate a vector, premultiply by this matrix.
    To rotate the coord sys underneath the vector, post multiply
    """
    return rotateInXMat(theta_deg)


def rotateAboutVectorMatrix(vec, theta_deg):
    """Construct the matrix that rotates vector a about
    vector vec by an angle of theta_deg degrees

    Taken from
    http://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle

    Input:
    theta_deg   (float) Angle through which vectors should be
                rotated in degrees

    Returns:
    A matrix

    To rotate a vector, premultiply by this matrix.
    To rotate the coord sys underneath the vector, post multiply

    """
    ct = np.cos(np.radians(theta_deg))
    st = np.sin(np.radians(theta_deg))

    # Ensure vector has normal length
    vec /= np.linalg.norm(vec)
    assert( np.all( np.isfinite(vec)))

    # compute the three terms
    term1 = ct * np.eye(3)

    ucross = np.zeros( (3,3))
    ucross[0] = [0, -vec[2], vec[1]]
    ucross[1] = [vec[2], 0, -vec[0]]
    ucross[2] = [-vec[1], vec[0], 0]

    term2 = st*ucross

    ufunny = np.zeros( (3,3))
    for i in range(0,3):
        for j in range(i,3):
            ufunny[i,j] = vec[i]*vec[j]
            ufunny[j,i] = ufunny[i,j]

    term3 = (1-ct) * ufunny

    return term1 + term2 + term3


def rotateInZMat(theta_deg):
    """Rotate a vector theta degrees around the z-axis

    Equivalent to yaw left

    Rotates the vector in the sense that the x-axis is rotated
    towards the y-axis. If looking along the z-axis (which is
    not the way you usually look at it), the vector rotates
    clockwise.

    If sitting on the vector [1,0,0], the rotation is towards the left

    Input:
    theta_deg   (float) Angle through which vectors should be
                rotated in degrees

    Returns:
    A matrix

    To rotate a vector, premultiply by this matrix.
    To rotate the coord sys underneath the vector, post multiply

    """

    ct = np.cos( np.radians(theta_deg))
    st = np.sin( np.radians(theta_deg))
    rMat = np.array([  [ ct, -st, 0],
                       [ st,  ct, 0],
                       [  0,   0, 1],
                    ])

    return rMat


def rotateInYMat(theta_deg):
    """Rotate a vector theta degrees around the y-axis

    Equivelent to pitch *down*

    Rotates the vector in the sense that the x-axis is rotated
    towards the negative y-axis. If looking along the y-axis,
    the vector rotates clockwise.

    If sitting on the vector [1,0,0], the rotation is down
    (towards the  south pole)

    Input:
    theta_deg   (float) Angle through which vectors should be
                rotated in degrees

    Returns:
    A matrix

    To rotate a vector, premultiply by this matrix.
    To rotate the coord sys underneath the vector, post multiply

    """
    ct = np.cos( np.radians(theta_deg))
    st = np.sin( np.radians(theta_deg))

    rMat = np.array([  [ ct,  0,  st],
                       [  0,  1,   0],
                       [-st,  0,  ct],
                    ])
    return rMat


def rotateInXMat(theta_deg):
    """Rotate a vector theta degrees around the x-axis

    Equivelent to rolling around x-axis clockwise

    Rotates the vector in the sense that the y-axis is rotated
    towards the z-axis. If looking along the x-axis,
    the vector rotates clockwise.

    If sitting on the vector [1,0,0] , the rotation is a clockwise
    roll.


    Input:
    theta_deg   (float) Angle through which vectors should be
                rotated in degrees

    Returns:
    A matrix

    To rotate a vector, premultiply by this matrix.
    To rotate the coord sys underneath the vector, post multiply
    """
    ct = np.cos( np.radians(theta_deg))
    st = np.sin( np.radians(theta_deg))

    rMat = np.array([  [  1,   0,  0],
                       [  0,  ct,  -st],
                       [  0, st,  ct],
                    ])
    return rMat


def vecFromRaDec(ra_deg, dec_deg):
    v = np.zeros((3,))

    ra_rad = np.radians(ra_deg)
    dec_rad = np.radians(dec_deg)
    ca = np.cos(ra_rad)
    sa = np.sin(ra_rad)
    cd = np.cos(dec_rad)
    sd = np.sin(dec_rad)

    v[0] = ca*cd
    v[1] = sa*cd
    v[2] = sd

    return v


def raDecFromVec(v):
    """
    Taken from
    http://www.math.montana.edu/frankw/ccp/multiworld/multipleIVP/spherical/learn.htm
    Search for "convert from Cartestion to spherical coordinates"

    Adapted because I'm dealing with declination which is defined
    with 90degrees at zenith

    @TODO: Rigourously test this against spice.recrad()
    """
    # Ensure v is a normal vector
    v /= np.linalg.norm(v)

    ra_deg = 0    # otherwise not in namespace0
    dec_rad = np.arcsin(v[2])
    s = np.hypot(v[0], v[1])
    if s == 0:
        ra_rad = 0
    else:
        ra_rad = np.arcsin(v[1]/s)
        ra_deg = np.degrees(ra_rad)
        if v[0] >= 0:
            if v[1] >= 0:
                pass
            else:
                ra_deg = 360 + ra_deg
        else:
            if v[1] > 0:
                ra_deg = 180 - ra_deg
            else:
                ra_deg = 180 - ra_deg

    raDec = ra_deg, np.degrees(dec_rad)
    return np.array(raDec)



"""
A set of functions to calculate properties of great circle lines on a
unit sphere, including, angular separation, bearing, etc.

Ported into Python from explanations at http://www.movable-type.co.uk/scripts/gis-faq-5.1.html
"""
import numpy as np


def sphericalAngSep(ra0, dec0, ra1, dec1, radians=False):
    """
        Compute the spherical angular separation between two
        points on the sky.

        //Taken from http://www.movable-type.co.uk/scripts/gis-faq-5.1.html

        NB: For small distances you can probably use
        sqrt( dDec**2 + cos^2(dec)*dRa)
        where dDec = dec1 - dec0 and
               dRa = ra1 - ra0
               and dec1 \approx dec \approx dec0
    """

    if radians==False:
        ra0  = np.radians(ra0)
        dec0 = np.radians(dec0)
        ra1  = np.radians(ra1)
        dec1 = np.radians(dec1)

    deltaRa= ra1-ra0
    deltaDec= dec1-dec0

    val = haversine(deltaDec)
    val += np.cos(dec0) * np.cos(dec1) * haversine(deltaRa)
    val = min(1, np.sqrt(val)) ; #Guard against round off error?
    val = 2*np.arcsin(val)

    #Convert back to degrees if necessary
    if radians==False:
        val = np.degrees(val)

    return val


def sphericalAngSepFast(ra0, dec0, ra1, dec1, radians=False):
    """A faster (but less accurate) implementation of sphericalAngleSep

    Taken from http://www.movable-type.co.uk/scripts/latlong.html

    For additional speed, set wantSquare=True, and the return value
    is the square of the separation
    """

    if radians==False:
        ra0  = np.radians(ra0)
        dec0 = np.radians(dec0)
        ra1  = np.radians(ra1)
        dec1 = np.radians(dec1)

    deltaRa= ra1-ra0
    deltaDec= dec1-dec0
    avgDec = .5*(dec0+dec1)

    x = deltaRa*np.cos(avgDec)
    val = np.hypot(x, deltaDec)

    if radians == False:
        val = np.degrees(val)

    return val


def haversine(x):
    """Return the haversine of an angle

    haversine(x) = sin(x/2)**2, where x is an angle in radians
    """
    y = .5*x
    y = np.sin(y)
    return y*y


def sphericalAngBearing(ra0, dec0, ra1, dec1, radians=False):
    sin = np.sin
    cos = np.cos
    atan  = np.arctan2

    if radians==False:
        ra0  = np.radians(ra0)
        dec0 = np.radians(dec0)
        ra1  = np.radians(ra1)
        dec1 = np.radians(dec1)

    dLong = ra1 - ra0
    a = sin(dLong)*cos(dec1)
    b = cos(dec1)*sin(dec1) - sin(dec0)*cos(dec1)*cos(dLong)
    bearing = atan(a, b)

    if radians==False:
        bearing = np.degrees(bearing)

    return bearing


def sphericalAngDestination(ra0_deg, dec0_deg, bearing_deg, dist_deg):
    sin = np.sin
    cos = np.cos
    asin  = np.arcsin
    atan2 = np.arctan2

    phi1 = np.radians(dec0_deg)    #Latitude
    lambda1 = np.radians(ra0_deg)    #Longitude
    d = np.radians(dist_deg)      #Distance in radians
    theta = np.radians(bearing_deg)

    phi2 = sin(phi1)*cos(d)
    phi2 += cos(phi1)*sin(d)*cos(theta)
    phi2 = asin(phi2)

    a = sin(theta)*sin(d)*cos(phi1)
    b = cos(d) - sin(phi1)*sin(phi2)
    lambda2 = lambda1 + atan2(a,b)

    ra2_deg = np.degrees(lambda2)
    dec2_deg = np.degrees(phi2)
    return ra2_deg, dec2_deg
